#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdarg.h>
#include <string.h>
#include <sys/wait.h>
#include <sys/poll.h>
#include <syslog.h>
#include "upsd.h"

int debug;
char *progname;
static int interval = 10;	/* secounds */
static const char *upsmonitor;	/* event-handler external program */

#define MSG_SYSLOG	0x01
#define MSG_CONSOLE	0x02
#define MSG_STDFILE	0x04

int msg_dest = MSG_STDFILE;

void vmsg(int flags, const char *fmt, va_list ap) {
  static const char warn[] = "warning";
  const char *err = (flags & MSG_ERRNO) && errno ? strerror(errno) : NULL;
  char *buf;
  int plen;
  int len = strlen(progname) + 2;
  if (flags & MSG_WARN) len += sizeof(warn) + 1;
  len += vsnprintf(NULL, 0, fmt, ap);
  if (err) len += strlen(err) + 2;
  buf = alloca(len + 2);
  plen = (flags & (MSG_ERROR|MSG_WARN)) ?  sprintf(buf, "%s: ", progname) : 0;
  if (flags & MSG_WARN)
    plen += sprintf(buf + plen, "%s: ", warn);
  len = plen;
  len += vsprintf(buf + len, fmt, ap);
  if (err)
    len += sprintf(buf + len, ": %s", err);
  if (msg_dest & MSG_SYSLOG) {
    int level =
      flags & MSG_ERROR ? LOG_ERR :
      flags & MSG_NOTICE ? LOG_NOTICE :
      flags & MSG_WARN  ? LOG_WARNING :
      flags & MSG_DEBUG ? LOG_DEBUG :
      LOG_INFO;
    char *m = buf + plen;
    buf[len] = '\0';
    openlog(progname, LOG_PID|LOG_NDELAY, LOG_USER);
    if (strchr(m, '%')) syslog(level, "%s", m);
    else syslog(level, m);
  }
  if (msg_dest & MSG_STDFILE) {
    buf[len] = '\n';
    write(flags & (MSG_ERROR|MSG_WARN) ? 1 : 2, buf, len + 1);
  }
  if (flags & MSG_FATAL)
    exit(2);
}

void msg(int flags, const char *fmt, ...) {
  va_list ap;
  va_start(ap, fmt); vmsg(flags, fmt, ap); va_end(ap);
}

void failed(const char *what) {
  eerr("%s failed", what);
}

static void oom(void) {
  msg(MSG_FATAL|MSG_ERROR, "out of memory");
}

void *emalloc(unsigned size) {
  void *ptr = malloc(size);
  if (!ptr) oom();
  return ptr;
}

void *ecalloc(unsigned size, unsigned count) {
  void *ptr = calloc(size, count);
  if (!ptr) oom();
  return ptr;
}

void *erealloc(void *ptr, unsigned size) {
  ptr = realloc(ptr, size);
  if (!ptr) oom();
  return ptr;
}

char *estrdup(const char *str) {
  char *dup = strdup(str);
  if (!dup) oom();
  return dup;
}

#define trealloc(type, ptr, count) \
	(type*)erealloc(ptr, sizeof(type) * (count))

static void tryhelp(const char *fmt, ...) {
  if (fmt) {
    va_list ap;
    va_start(ap, fmt);
    vmsg(MSG_ERROR, fmt, ap);
    va_end(ap);
  }
  msg(MSG_ERROR|MSG_FATAL, "try `%s -h' for help", progname);
}

static void run_action(const struct ups *ups, const char *action) {
  int status;
  if (!upsmonitor)
    return;
  dbg(1, "executing %s %s %s", upsmonitor, action, ups->tag);
  if (fork() == 0) {
    execl(upsmonitor, upsmonitor, action, ups->tag, NULL);
    ewarn("unable to execute %s", upsmonitor);
    exit(255);
  }
  wait(&status);
}

static struct ups_stat_bits {
  unsigned bit;
  const char *name;
} statbits[] = {
  { UPSS_ONBATT, "onbatt" },
  { UPSS_LOWBATT, "lowbatt" },
  { UPSS_OVERLOAD, "overload" },
  { UPSS_BATTFAIL, "battfail" },
  { UPSS_BOOST, "avrboost" },
  { UPSS_BUCK, "avrbuck" },
  { UPSS_FAILURE, "upsfail" },
  { UPSS_SELFTEST, "selftest" },
  { UPSS_OFF, "upsoff" },
};

static void pstatus(int status) {
  fputs("status bits:", stdout);
  if (status < 0) puts(status == UPSR_PROTO ? " proto" : " ioerr");
  else if (!status) puts(" (none)");
  else {
    unsigned i;
    printf(" 0x%x", status);
    for(i = 0; i < sizeof(statbits)/sizeof(statbits[0]); ++i)
      if (status & statbits[i].bit)
        printf(" %s", statbits[i].name);
    putchar('\n');
  }
}

static void pcinfo(const struct ups_cinfo *cinfo) {
  printf("info:");
#define pr(x) \
    if (cinfo->x >= 0) \
      printf(" %s=%d.%d", #x, \
        cinfo->x / 1000, \
	(cinfo->x % 1000)/100)
  pr(ivolt); pr(ovolt);
  pr(ifreq); pr(ofreq);
  pr(loadpct); pr(battpct);
#undef pr
  putchar('\n');
}

static int init(int argc, char **argv, struct ups *ups) {
  int r, status;
  const struct ups_driver **drvp, *drv;
  int do_probe = 0;
  int do_cmd = 0;
  int do_fork = 1;

  if ((progname = strrchr(argv[0], '/')) != NULL) argv[0] = ++progname;
  else progname = argv[0];

  while((r = getopt(argc, argv, "a:i:c:nfvphH:")) != EOF) switch(r) {
  case 'a': upsmonitor = optarg; break;
  case 'p': do_probe = 1; break;
  case 'c':
    if (strcmp(optarg, "restart") == 0) do_cmd = UPSC_RESTART;
    else if (strcmp(optarg, "halt") == 0) do_cmd = UPSC_HALT;
    else if (strcmp(optarg, "wake") == 0) do_cmd = UPSC_WAKE;
    else if (strcmp(optarg, "selftest") == 0) do_cmd = UPSC_SELFTEST;
    else tryhelp("unknown command (-c) `%s'", optarg);
    break;
  case 'f': do_fork = 0; break;
  case 'n': do_fork = 0; break;
  case 'i':
    interval = atoi(optarg);
    if (interval < 1)
      tryhelp("invalid interval (-i) `%s'", optarg);
    break;
  case 'v': ++debug; break;
  case 'h':
    printf(
"%s: UPS monitor daemon.  Usage is:\n"
" %s [options] tag:driver:device:name=value,name=value... ...\n"
"where:\n"
" tag is a string identifier for the ups, such as ups0,\n"
" driver is the UPS driver to use,\n"
" device is a driver-specific device name (usually /dev/something)\n"
" name=value is an optional list of driver-specific parameters\n"
"and options are:\n"
" -i interval - polling interval in secounds\n"
" -c cmd[=arg] - execute a driver-specific command `cmd'.\n"
"  Commonly recognized commands are:\n"
"  shutdown[=sec] order UPS to shut down in a few secounds\n"
"\tand return online when the power will be back\n"
"  halt[=sec] - order UPS to shut down\n"
"  wake - wake up the UPS if turned off\n"
"  selftest[=sec] - request selftest (for n seconds)\n"
" -p - probe UPS only\n"
" -n - don't fork, run in foreground\n"
" -a pathname - run this program to report events\n"
"\t(specify empty string or literal string \"none\" to disable)\n"
" -v - increase verbosity/debug level\n"
" -h - print this help and exit\n"
" -H driver - print driver-specific info and exit\n"
, progname, progname);
    printf("The following UPS drivers are available:\n");
    for(drvp = drvtab; (drv = *drvp); ++drvp)
      printf(" %s %s\n", drv->name, drv->description);
    exit(0);
  case 'H':
    for(drvp = drvtab; (drv = *drvp); ++drvp)
      if (streq(drv->name, optarg)) {
        printf("%s: %s\n%s", optarg, drv->description, drv->help);
        exit(0);
      }
    tryhelp("unknown driver `%s'", optarg);
  default: tryhelp(0);
  }

  argc -= optind; argv += optind;
  if (!argc)
    tryhelp("no ups specification given");

  memset(ups, 0, sizeof(*ups));
  ups->fd = -1;
  memset(&ups->cinfo, 0xff, sizeof(ups->cinfo));

  for(r = 0; r < /*argc*/1; ++r) {
    char *s = estrdup(argv[r]);
    char *t;
    char **av;
    int n;
    for(n = 0, t = strtok(s, ","); t; t = strtok(NULL, ","))
      ++n;
    av = alloca(sizeof(*av) * (n + 1));
    strcpy(s, argv[r]);
    for(n = 0, t = strtok(s, ","); t; t = strtok(NULL, ",")) {
           if (strpfx(t, "tag=")) ups->tag    = t + 4;
      else if (strpfx(t, "dev=")) ups->device = t + 4;
      else if (strpfx(t, "drv=")) ups->driver = t + 4;
      else if (strpfx(t, "driver=")) ups->driver = t + 7;
      else if (strpfx(t, "device=")) ups->device = t + 7;
      else av[n++] = t;
    }
    av[n] = NULL;
    if (!ups->tag) {
      char b[3+10+1];
      sprintf(b, "ups%d", r);
      ups->tag = estrdup(b);
    }
    if (!ups->driver)
      tryhelp("no driver specified for %s in `%s'", ups->tag, argv[r]);
    if (!ups->device)
      tryhelp("no device specified for %s in `%s'", ups->tag, argv[r]);
     
    for(drvp = drvtab; (drv = *drvp); ++drvp)
      if (streq(drv->name, ups->driver))
        break;
    if (!drv)
      tryhelp("unknown model/driver `%s' for %s", ups->driver, ups->tag);
    ups->openfn = drv->openfn;
    dbg(1, "%s: opening %s with %s driver", ups->tag, ups->device, ups->driver);
    status = ups->openfn(ups, av);
    if (status < 0) {
      if (status == UPSR_PROTO)
        fatal("unable to communicate with UPS");
      exit(1);
    }
    dbg(1, "%s: found %s UPS on %s (%s)", ups->tag, ups->driver, ups->device,
        ups->model ? ups->model : "model unavailable");
    if (debug) {
      pstatus(status);
      if (ups->flags & UPSF_CINFO) pcinfo(&ups->cinfo);
    }

  }

  if (do_probe || do_cmd) {
    upsmonitor = NULL;
    do_fork = 0;
  }

  if (do_probe) {
    dbg(1, "probe succeseful\n");
    exit(0);
  }
  if (do_cmd) {
    r = ups->cmdfn(ups, do_cmd);
    exit(r > 0 ? 0 : 2);
  }
  if (do_fork) {
    if (fork())
      exit(0);
    chdir("/");
    close(0);
    open("/dev/null", O_RDWR);
    dup2(0, 1);
    dup2(0, 2);
    setsid();
  }

  return status;
}

#define bit_set(status, bit) ((status) & (bit))

static void report_status_change(struct ups *ups, int os, int s) {

  if (s < 0) {
    run_action(ups, s == UPSR_PROTO ? "noconn" : "ioerr");
    return;
  }
  else if (os < 0)
    run_action(ups, "conn");

  if (bit_set(s, UPSS_ONBATT)) {
    if (!bit_set(os, UPSS_ONBATT))
      run_action(ups, "onbatt");
    if (bit_set(s, UPSS_LOWBATT|UPSS_OVERLOAD|UPSS_BATTFAIL|UPSS_FAILURE))
      run_action(ups, "lowbatt");
  }
  else if (os & UPSS_ONBATT)
    run_action(ups, "online");

#define simple_action(flag, name) if (bit_set(s, flag)) run_action(ups, name)
#define change_action(flag, name) \
    if (bit_set(s ^ os, flag)) \
      run_action(ups, bit_set(s, flag) ? name : name "-end")

  simple_action(UPSS_OVERLOAD, "overload");
  simple_action(UPSS_BATTFAIL, "battfail");
  simple_action(UPSS_FAILURE, "upsfail");
  change_action(UPSS_SELFTEST, "selftest");
  change_action(UPSS_BOOST, "avrboost");
  change_action(UPSS_BUCK, "avrbuck");

#undef simple_action
#undef change_action

}

int main(int argc, char **argv) {
  struct ups ups;
  int ostatus, status;

  ostatus = init(argc, argv, &ups);
  run_action(&ups, "started");

  status = 0;

  for(;;) {
    if (status != 0 || ostatus != status) {
      report_status_change(&ups, ostatus, status);
      ostatus = status;
    }
    sleep(interval);
    if (ups.flags & UPSF_CINFO)
      memset(&ups.cinfo, 0xff, sizeof(ups.cinfo));
    status = ups.pollfn(&ups);
    if (debug) {
      pstatus(status);
      if (ups.flags & UPSF_CINFO) pcinfo(&ups.cinfo);
    }
  }

  return 0;
}

static void
showdata(const struct ups *ups, const char *p, const unsigned char *b, int len)
{
  int r, c;
  printf("%s%d:", p, len);
  if (ups->flags & UPSF_BINARY) {
    for(r = 0; r < len; ++r)
      printf(" 0x%02x(%u)", b[r], b[r]);
  }
  else {
    putchar(' ');
    for(r = 0; r < len; ++r) {
      switch(b[r]) {
      case '\\': c = '\\'; break;
      case '\n': c = 'n'; break;
      case '\t': c = 't'; break;
      case '\r': c = 'r'; break;
      case '\v': c = 'v'; break;
      default:
        if (b[r] >= ' ' && b[r] < 0x7f) putchar(b[r]);
        else printf("\\%03o", b[r]);
        continue;
      }
      printf("\\%c", c);
    }
  }
  putchar('\n');
}

int ups_write(struct ups *ups, const void *wbuf, int wlen) {
  int r;
  const unsigned char *wb = wbuf;
  if (debug > 2)
    showdata(ups, "writing ", wb, wlen);
  do {
    r = write(ups->fd, wb, wlen);
    if (r <= 0) {
      if (r < 0)
        eerr("ups %s on %s failed", "write()", ups->device);
      else
        err("ups write() on %s: short write", ups->device);
      return UPSR_ERROR;
    }
    wb += r;
    wlen -= r;
  } while(wlen);
  return 0;
}

int ups_read(struct ups *ups, void *rbuf, int rlen, int mstmo) {
  unsigned char *rb = rbuf;
  struct pollfd pfd;
  int r;
  pfd.fd = ups->fd;
  pfd.events = POLLIN;
  pfd.revents = 0;
  r = poll(&pfd, 1, mstmo);
  if (r < 0)
    efatal("poll");
  else if (!r) {
    dbg(3, "timeout reading from UPS");
    rb[0] = '\0';
    return 0;
  }
  else if ((r = read(ups->fd, rbuf, rlen)) <= 0) {
    rb[0] = '\0';
    if (r < 0) {
      eerr("ups read failed");
      return UPSR_ERROR;
    }
    else {
      dbg(3, "eof reading from UPS");
      return 0;
    }
  }
  else if (debug > 2)
    showdata(ups, "reading ", rb, r);
  ups->nread += r;
  if (r < rlen)
    rb[r] = '\0';
  return r;
}
