#include <sys/types.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <stdio.h>

#ifdef __GNUC__
# define UNUSED __attribute__((unused))
# define PRINTFLIKE(fmt,arg) __attribute__((format(printf,fmt,arg)))
# define NORETURN __attribute__((noreturn))
#else
# define UNUSED
# define PRINTFLIKE(fmt,arg)
# define NORETURN
#endif

void *emalloc(unsigned size);
void *ecalloc(unsigned size, unsigned count);
#define tmalloc(type, count) (type*)emalloc(sizeof(type)*count)
#define tcalloc(type, count) (type*)ecalloc(sizeof(type), count)
char *estrdup(const char *str);

#define streq(s,a) (strcmp((s),(a))==0)
#define streq2(s,a,b) (streq((s),(a))||streq((s),(b)))
#define streq3(s,a,b,c) (streq2((s),(a),(b))||streq((s),(c)))
#define streq4(s,a,b,c,d) (streq3((s),(a),(b),(c))||streq((s),(d)))

#define strpfx(str, pfx) (strncmp(str,pfx,sizeof(pfx)-1)==0)

extern int debug;
void PRINTFLIKE(2,3) msg(int flags, const char *fmt, ...);
#define MSG_ERRNO	0x01
#define MSG_ERROR	0x02
#define MSG_WARN	0x04
#define MSG_DEBUG	0x08
#define MSG_FATAL	0x10
#define MSG_NOTICE	0x20
#define warn(...)	msg(MSG_WARN, __VA_ARGS__)
#define ewarn(...)	msg(MSG_WARN|MSG_ERRNO, __VA_ARGS__)
#define err(...)	msg(MSG_ERROR, __VA_ARGS__)
#define eerr(...)	msg(MSG_ERROR|MSG_ERRNO, __VA_ARGS__)
#define fatal(...)	msg(MSG_FATAL, __VA_ARGS__)
#define efatal(...)	msg(MSG_FATAL|MSG_ERRNO, __VA_ARGS__)
void failed(const char *what);
#define dbg(level,...) \
	do { if (debug >= level) msg(MSG_DEBUG,__VA_ARGS__); } while(0)

struct ups;
struct ups_driver_data;

#define UPS_DEFAULT_SHUTDOWN_DELAY 5	/* secounds */

/* commands for ups_cmd_fn() */
enum ups_cmd {
  UPSC_RESTART = 1,	/* shutdown and restart when back on power */
  UPSC_HALT,		/* turn UPS off */
  UPSC_WAKE,		/* turn UPS on */
  UPSC_SELFTEST		/* selftest command */
};

/* ups_open_fn() and ups_poll_fn() return:
 *  -1 - generic i/o error while communicating with ups
 *  -2 - unexpected/no response from ups
 *  >=0 - UPSS_XXX bits */
typedef int ups_open_fn(struct ups *ups, char **argv);
typedef int ups_poll_fn(struct ups *ups);
/* ups_init_fn initializes some structure members and allocates private data */
typedef int ups_init_fn(struct ups *ups);

typedef int ups_cmd_fn(struct ups *ups, enum ups_cmd cmd);

struct ups_cinfo {
  int status;		/* UPSS_XXX bits - general UPS status */
  int ivolt, ovolt;	/* volts/1000, millivolts */
  int ifreq, ofreq;	/* hertz/1000, millihertz */
  int loadpct;		/* UPS load percentage /1000, millipercents */
  int battpct;		/* battery charge percentage /1000, millipercents */
};

struct ups {
  char *tag;			/* identification tag */
  char *device;			/* name of the device */
  char *name;			/* name of the UPS */
  char *model;			/* model name/specification */
  const char *driver;		/* driver name */
  int flags;			/* device flags, UPSF_XXX */
  struct ups_cinfo cinfo, ocinfo; /* additional info, cur and previous */
  ups_open_fn *openfn;		/* re-open routine */
  ups_poll_fn *pollfn;		/* routine to poll status */
  ups_cmd_fn *cmdfn;		/* command routine */

  int fd;			/* filedescriptor */
  struct ups_driver_data *data;	/* driver-specific data */
  int ds[6];			/* driver-specific data */
  int nread;			/* number of bytes read from fd */
  struct ups *next;
};

/* status bits */
#define UPSS_ONBATT	0x0001	/* UPS is on battery */
#define UPSS_LOWBATT	0x0002	/* the battery is low */
#define UPSS_OVERLOAD	0x0004	/* UPS is overloaded */
#define UPSS_BATTFAIL	0x0008	/* the battery failed */
#define UPSS_FAILURE	0x0010	/* UPS indicates some failure */
#define UPSS_SELFTEST	0x0020	/* UPS selftest is in progress */
#define UPSS_BOOST	0x0040  /* currently upping the input voltage */
#define UPSS_BUCK	0x0080	/* currently lowering the input voltage */
#define UPSS_OFF	0x0100	/* UPS is turned OFF */

#define UPSR_ERROR	-1	/* I/O error communicating with UPS */
#define UPSR_PROTO	-2	/* UPS response/protocol mismatch */

/* flag bits.
 * only the high byte is used, the rest is available to drivers */
#define UPSF_BINARY	0x01000000	/* show commands in binary form */
#define UPSF_CINFO	0x02000000	/* gather extended info (cinfo) */

struct ups_driver {
  const char *name;
  const char *description;
  const char *help;
  ups_open_fn *openfn;
//  ups_init_fn *initfn;
};

extern const struct ups_driver *drvtab[];

#define UPS_DRIVER(name,pfx,descr) \
 static ups_open_fn pfx##_open; \
 static ups_init_fn pfx##_init; \
 const struct ups_driver _##pfx##_drv = { name, descr, pfx##_help, pfx##_open };

int ups_write(struct ups *ups, const void *wbuf, int wlen);
int ups_read(struct ups *ups, void *rbuf, int rlen, int mstmo);
